/* DIFFH - A diff frontend for output in HTML format.
 * Module:  CHUNK
 * Author:  Matt Giuca
 * Created: 5/2/06
 * Updated: 15/4/06
 *
 * Copyright (C) 2006 Matt Giuca <mattgiuca@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either 
 * version 2 of the License, or (at your option) any later 
 * version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* Module: CHUNK
 * This module forms formal chunk structures from chunk lines, parses them,
 * and writes them to the output in HTML format.
 */

#include <ctype.h>
#include "global.h"
#include "chunk.h"

/* Character types where word parsing is concerned */
#define CHAR_OTHER 0
#define CHAR_ALPHANUM 1
#define CHAR_SPACE 2

/* In find_changes, b_mat values */
#define BMAT_0 0
#define BMAT_UL 1
#define BMAT_U 2
#define BMAT_L 3

/* Determines which characters are "alphanumeric" and therefore form a single
    word of contiguous */
/* Current definition: c is alphanumeric if it is any of the following:
    - A lowercase or uppercase letter of the alphabet
    - A digit
    - An apostrophe (')
*/
/* Checks all Unicode characters which resemble Latin letters */
#define IS_ALPHANUM(c) isalnum(c & 0xFF) || c == '\'' \
    || c == 0x8A || c == 0x8E || c == 0x9E || c == 0x9F \
    || (c >= 0xC0 && c <= 0xD6) || (c >= 0xD8 && c <= 0xDD) \
    || (c >= 0xE0 && c <= 0xF6) || (c >= 0xF9 && c <= 0xFD) \
    || (c >= 0xFF && c <= 0x17F) || (c >= 0x1C4 && c <= 0x1FD)

/* PRIVATE DATA TYPES */

/* Note that Chunk only stores one half-chunk */
typedef struct {
    int* line_index;    /* Start word indices for each line */
    char** word;        /* Strings containing the words */
    char* word_changed; /* boolean - corresponds to word */
    int numlines;
    int numwords;
} Chunk;

/* PRIVATE FUNCTION DECLARATIONS */
static Chunk* lines_to_chunk(int numlines, char** lines);
static void free_chunk(Chunk* ch);
static void free_lines(int numlines, char** lines);
static void find_changes(Chunk* deleted, Chunk* added);
static char** chunk_to_html(Chunk* chunk, int side);
static void interlace_and_print_html(int numdeleted, char** deleted,
                                     int numadded, char** added);
static unsigned int get_unicode(unsigned char** ptr);

/* MODULE FUNCTION CODE */

/* Function: chunk_process
 * Takes two string arrays, the lines on either side of the chunk.
 * Fully processes the chunks and outputs them to HTML.
 * Precond: Global outstream must be valid.
 */
void chunk_process(int numdeleted, char** deleted,
                   int numadded, char** added)
{
    Chunk* ch_deleted, * ch_added;
    char** html_deleted, ** html_added;

    /* 1. Convert the arrays of strings into formal half-Chunks */
    ch_deleted = lines_to_chunk(numdeleted, deleted);
    ch_added = lines_to_chunk(numadded, added);

    /* 2. Detect any differences between the two half-chunks */
    if (numdeleted && numadded)             /* Only if 2-sided */
        find_changes(ch_deleted, ch_added); /* Sets word_changed arrays */

    /* 3. Parse each half-chunk into HTML - individual table cells */
    html_deleted = chunk_to_html(ch_deleted, 0);
    html_added = chunk_to_html(ch_added, 1);

    /* Free the Chunk structures created in (1) */
    free_chunk(ch_deleted);
    free_chunk(ch_added);

    /* 4. Interlace the table cells and write to the output stream */
    /* Note that html_X should correspond precisely to argument X */
    interlace_and_print_html(numdeleted, html_deleted, numadded, html_added);

    /* Free the html lines */
    free_lines(numdeleted, html_deleted);
    free_lines(numadded, html_added);
}

/* Function: lines_to_chunk
 * Takes a string array and parses it into a formal Chunk structure.
 * The word_changed array is set to all 0s, it is not computed here.
 * Returns: Pointer to the Chunk structure containing the text of lines.
 * This structure contains allocated memory and must be freed with free_chunk
 */
Chunk* lines_to_chunk(int numlines, char** lines)
{
    int i;
    char* j, * oldj;
    unsigned int ucchar;
    int currenttype;        /* CHAR_OTHER/ALPHANUM/SPACE */
    int currentwordidx = 0, first_char_flag;
    int word_malloc_size = DEFAULT_MALLOC;
    char* currentword;      /* Points to an element of lines */
    char* newword;          /* Newly malloc'd for storage */

    Chunk* ch = malloc(sizeof(Chunk));
    ch->line_index = malloc(numlines * sizeof(int));
    ch->numlines = numlines;
    ch->word = malloc(DEFAULT_MALLOC * sizeof(char*));

    for (i=0; i<numlines; i++)
    {
        /* Start new line and break current word */
        ch->line_index[i] = currentwordidx;
        currenttype = CHAR_OTHER;
        first_char_flag = FALSE;
        /* Only loop if line is not empty */
        if (*lines[i])
        {
            for (j=lines[i]; *j; j++)
            {
                /* Grab the whole Unicode char into word, and
                    work out its numeric code */
                oldj = j;
                ucchar = get_unicode((unsigned char**) &j);
                /* Decide whether to terminate word or keep reading */
                if ((currenttype == CHAR_ALPHANUM && (IS_ALPHANUM(ucchar)))
                    || (currenttype == CHAR_SPACE && ucchar == ' '))
                { /* Keep reading this word */ }
                else
                {   /* Terminate word */
                    if (first_char_flag)
                    {   /* Not 1st character of line - store current word */
                        newword = malloc((oldj-currentword+1) *sizeof(char));
                        strncpy(newword, currentword, oldj - currentword);
                        newword[oldj-currentword] = '\0';
                        if (currentwordidx >= word_malloc_size)
                            ch->word = realloc(ch->word,
                                (word_malloc_size*=2) * sizeof(char*));
                        ch->word[currentwordidx++] = newword;
                    }
                    first_char_flag = TRUE;
                    /* Start new word with j */
                    currentword = oldj;
                    /* If alphanum or space, set that type */
                    if (IS_ALPHANUM(ucchar))
                        currenttype = CHAR_ALPHANUM;
                    else if (ucchar == ' ')
                        currenttype = CHAR_SPACE;
                    else
                        currenttype = CHAR_OTHER;
                }
            }
            /* Terminate final word for line */
            newword = malloc((j - currentword + 1) * sizeof(char));
            strncpy(newword, currentword, j - currentword);
            newword[j-currentword] = '\0';
            if (currentwordidx >= word_malloc_size)
                ch->word = realloc(ch->word,
                    (word_malloc_size*=2) * sizeof(char*));
            ch->word[currentwordidx++] = newword;
        }
        else    /* The whole line is empty - make blank word for this line */
        {
            newword = malloc(sizeof(char));
            *newword = '\0';
            if (currentwordidx >= word_malloc_size)
                ch->word = realloc(ch->word, (word_malloc_size*=2)
                    * sizeof(char*));
            ch->word[currentwordidx++] = newword;
        }
    }
    ch->numwords = currentwordidx;

    /* Allocate word_changed - one per word, and initialise to 0 */
    ch->word_changed = malloc(currentwordidx * sizeof(char));
    for (i=0; i<currentwordidx; i++)
        ch->word_changed[i] = FALSE;

    return ch;
}

/* Function: free_chunk
 * Frees up all memory associated with a chunk, including the chunk itself.
 */
void free_chunk(Chunk* chunk)
{
    int i;
    free(chunk->line_index);
    free(chunk->word_changed);
    for (i=0; i<chunk->numwords; i++)
        free(chunk->word[i]);
    free(chunk->word);
    free(chunk);
}

/* Function: find_changes
 * Takes two (half) Chunks and detects the differences between them.
 * Side-effects: Sets the word_changed arrays of both chunks to reflect their
 * differences.
 */
void find_changes_(Chunk* deleted, int d_start, int d_end,
                   Chunk* added, int a_start, int a_end)
{
    int i, j;
    char t;
    char** d_word = deleted->word;
    char** a_word = added->word;
    int lcs_d_start, lcs_d_end, lcs_a_start, lcs_a_end, lcs_len=0;
    int cur_string_d_end, cur_string_a_end, cur_string_len;

    /* LCS matrices */
    int mx_width = d_end - (--d_start) + 1;
    int mx_height = a_end - (--a_start) + 1;
        /* Note: d_start and a_start have been decremented here */
    short* c_mat = malloc(mx_width * mx_height * sizeof(short));
    char* b_mat = malloc(mx_width * mx_height * sizeof(char));

    if (mx_width <= 1 || mx_height <= 1)    /* Timesaver for base case */
    {   /* Mark all words as changed and exit */
        for (i=d_start+1; i<=d_end; i++)
            deleted->word_changed[i] = TRUE;
        for (i=a_start+1; i<=a_end; i++)
            added->word_changed[i] = TRUE;
        return;
    }

    /* Find the Longest Common Subsequence of unchanged words */
    /* Uses O(d*a) algorithm from
        http://en.wikipedia.org/wiki/Longest-common_subsequence_problem */
    for (i=0; i<mx_height; i++)
    {
        c_mat[i*mx_width] = 0;
        b_mat[i*mx_width] = 0;
    }
    for (j=0; j<mx_width; j++)
    {
        c_mat[j] = 0;
        b_mat[j] = 0;
    }

    /* i+a_start corresponds to a_word, j+d_start corresponds to d_word */
    for (i=1; i<mx_height; i++)
        for (j=1; j<mx_width; j++)
        {
            if (!strcmp(d_word[j+d_start], a_word[i+a_start]))
            {
                c_mat[i*mx_width + j] = c_mat[(i-1)*mx_width + j-1] + 1;
                b_mat[i*mx_width + j] = BMAT_UL;
            }
            else if (c_mat[(i-1)*mx_width + j] >= c_mat[i*mx_width + j-1])
            {
                c_mat[i*mx_width + j] = c_mat[(i-1)*mx_width + j];
                b_mat[i*mx_width + j] = BMAT_U;
            }
            else {
                c_mat[i*mx_width + j] = c_mat[i*mx_width + j-1];
                b_mat[i*mx_width + j] = BMAT_L;
            }
        }

    /* Read off the b matrix and search for the longest continuous sequence
        of BMAT_ULs */
    i--; j--;   /* Bring back onto the grid - bottom-right corner */
    t = b_mat[i*mx_width + j];
    cur_string_len = 0;
    while (t != BMAT_0)
    {
        if (t == BMAT_UL)   /* Go up-left and keep/start sequence */
        {
            if (cur_string_len == 0)
            {
                cur_string_d_end = j+d_start;
                cur_string_a_end = i+a_start;
            }
            i--; j--;
            cur_string_len++;
        }
        else                /* Break sequence */
        {
            if (cur_string_len >= lcs_len)
            {
                lcs_d_start = j+d_start+1;
                lcs_d_end = cur_string_d_end;
                lcs_a_start = i+a_start+1;
                lcs_a_end = cur_string_a_end;
                lcs_len = cur_string_len;
            }
            cur_string_len = 0;
            /* Go up or left */
            if (t == BMAT_U)
                i--;
            else
                j--;
        }
        t = b_mat[i*mx_width + j];
    }
    /* Break last sequence */
    if (cur_string_len >= lcs_len)
    {
        lcs_d_start = j+d_start+1;
        lcs_d_end = cur_string_d_end;
        lcs_a_start = i+a_start+1;
        lcs_a_end = cur_string_a_end;
        lcs_len = cur_string_len;
    }

    /* Chaos clause - if LCS length is 1, and that word is a space,
        treat as not having found any matching words.
        (Since spaces don't show up as changes anyway) */
    if (lcs_len == 1 && *d_word[lcs_d_start] == ' ') lcs_len = 0;

    /* DEBUG OUTPUT - print out the matrices */
#ifdef PRINTDEBUG
    printf("FIND_CHANGES MATRIX:\n%d x %d\n- ", mx_width, mx_height);
    for (i=d_start+1; i<=d_end; i++)
        printf("\"%s\" ", d_word[i]);
    printf("\n+ ");
    for (i=a_start+1; i<=a_end; i++)
        printf("\"%s\" ", a_word[i]);
    putchar('\n');
    for (i=0; i<mx_height; i++)
    {
        for (j=0; j<mx_width; j++)
            printf("%d\t", c_mat[i*mx_width + j]);
        putchar('\n');
    }
    putchar('\n');
    for (i=0; i<mx_height; i++)
    {
        for (j=0; j<mx_width; j++)
            switch (b_mat[i*mx_width + j])
            {
                case BMAT_0:
                printf("0\t");
                break;
                case BMAT_UL:
                printf("UL\t");
                break;
                case BMAT_U:
                printf("U\t");
                break;
                default:
                printf("L\t");
            }
        putchar('\n');
    }
    if (lcs_len)
    {
        printf("Longest sequence: (%d,%d)-(%d,%d) = %d\n", lcs_d_start,
            lcs_a_start, lcs_d_end, lcs_a_end, lcs_len);
        for (i=lcs_d_start; i<=lcs_d_end; i++)
            printf("\"%s\" ", d_word[i]);
        putchar('\n');
    }
    else
        printf("No matching words.\n");

    putchar('\n');
#endif
    /* END DEBUG OUTPUT */

    /* Free matrices */
    free(c_mat);
    free(b_mat);

    /* Recursive decision */
    /* Keep in mind here that a_start and d_start are -1 than they should
        be - ie. how they were passed as arguments */
    if (lcs_len)
    {   /* Recurse, on either side of the LCS */
        find_changes_(deleted, d_start+1, lcs_d_start-1, added, a_start+1, lcs_a_start-1);
        find_changes_(deleted, lcs_d_end+1, d_end, added, lcs_a_end+1, a_end);
    }
    else
    {   /* Base case - nothing in common. Mark all words changed */
        for (i=d_start+1; i<=d_end; i++)
            deleted->word_changed[i] = TRUE;
        for (i=a_start+1; i<=a_end; i++)
            added->word_changed[i] = TRUE;
    }
}
void find_changes(Chunk* deleted, Chunk* added)
{
    find_changes_(deleted, 0, deleted->numwords-1, added, 0,
        added->numwords-1);
}

/* Function chunk_to_html
 * Takes one (half) Chunk and formats it into HTML text, using <span> tags
 * to show highlighting where necessary.
 * side = 0 for "deleted" side, 1 for "added" side
 * Returns: An array of strings containing the HTML code, one string per line
 * Note: This will convert characters to HTML escapes where necessary.
 */
char** chunk_to_html(Chunk* chunk, int side)
{
    int i, j = 0;
    int t, numspaces, numspaces_d2;
    char highlight_on = 0;  /* boolean */
    char* highlight_text;
    int numlines = chunk->numlines;
    char** lines = malloc(numlines * sizeof(char*));
    char* thisline = malloc(DEFAULT_MALLOC_BIG * sizeof(char));
    size_t thislinelen = 0, thislinelen_new = 0;
    unsigned int thisline_malloc_size = DEFAULT_MALLOC_BIG;
    char* thisword;
    char thisword_mallocd;  /* boolean */

    /* DEBUG OUTPUT - print the whole Chunk */
#ifdef PRINTDEBUG
    printf("CHUNK:\n");
    for (i=0; i<chunk->numwords; i++)
    {
        if (chunk->line_index[j] == i)
            printf("LINE #%d BEGINS\n", j++);
        printf("\"%s\"\t%d\n", chunk->word[i], chunk->word_changed[i]);
    }
    j = 0;
    putchar('\n');
#endif

    *thisline = '\0';
    for (i=0; i<chunk->numwords; i++)
    {
        if (j+1 < numlines && chunk->line_index[j+1] == i)
        {
            /* Close highlight, expanding thisline if necessary */
            if (highlight_on)
            {
                thislinelen_new += strlen(HIGHLIGHT_OFF);
                if (thislinelen_new+1 >= thisline_malloc_size)
                    thisline = realloc(thisline, (thisline_malloc_size*=2)
                        * sizeof(char));
                strcpy(thisline+thislinelen, HIGHLIGHT_OFF);
                highlight_on = FALSE;
            }
            lines[j++] = thisline;
            thisline = malloc(DEFAULT_MALLOC_BIG * sizeof(char));
            thisline_malloc_size = DEFAULT_MALLOC_BIG;
            *thisline = '\0';
            thislinelen = thislinelen_new = 0;
        }

        /* Apply changes to highlight state */
        highlight_text = NULL;
        if (highlight_on)
        {
            if (!chunk->word_changed[i])
            {
                highlight_text = HIGHLIGHT_OFF;
                highlight_on = FALSE;
            }
        }
        else
        {
            if (chunk->word_changed[i])
            {
                highlight_text = side ?
                    HIGHLIGHT_ON_ADDED : HIGHLIGHT_ON_DELETED;
                highlight_on = TRUE;
            }
        }

        /* Copy word to thisword, changing it if necessary to valid HTML */
        thisword = chunk->word[i];
        thisword_mallocd = FALSE;
        switch (*thisword)
        {
            case ' ':   /* Spaces - if multiple, replace with &nbsp; */
            case '\t':
            if (*thisword == ' ')
            {
                numspaces = (int) strlen(thisword);
                if (numspaces == 1) break;
            }
            else
                numspaces = tabwidth;
            numspaces_d2 = numspaces / 2;
            thisword = malloc((7*numspaces_d2 + 6*(numspaces%2) + 1)
                * sizeof(char));
            for (t=0; t<numspaces_d2; t++)
                strncpy(thisword+(t*7), HT_NBSP " ", 7);
            if (numspaces%2) strncpy(thisword+(t*7), HT_NBSP, 6);
            thisword[7*(numspaces/2) + 6*(numspaces%2)] = '\0';
            thisword_mallocd = TRUE;
            break;
            case '&':
            thisword = HT_AMP;
            break;
            case '\"':
            thisword = HT_QUOT;
            break;
            case '<':
            thisword = HT_LT;
            break;
            case '>':
            thisword = HT_GT;
            break;
        }

        /* Expand thisline if necessary */
        if (highlight_text) thislinelen_new += strlen(highlight_text);
        thislinelen_new += strlen(thisword);
        if (thislinelen_new+1 >= thisline_malloc_size)
            thisline = realloc(thisline, (thisline_malloc_size*=2)
                * sizeof(char));
        if (highlight_text)
        {
            strcpy(thisline+thislinelen, highlight_text);
            strcpy(thisline+thislinelen+strlen(highlight_text), thisword);
        }
        else
            strcpy(thisline+thislinelen, thisword);
        if (thisword_mallocd) free(thisword);
        thislinelen = thislinelen_new;
    }
    /* Close highlight, expanding thisline if necessary */
    if (highlight_on)
    {
        thislinelen_new += strlen(HIGHLIGHT_OFF);
        if (thislinelen_new+1 >= thisline_malloc_size)
            thisline = realloc(thisline, (thisline_malloc_size*=2)
                * sizeof(char));
        strcpy(thisline+thislinelen, HIGHLIGHT_OFF);
    }
    lines[j] = thisline;

    /* DEBUG OUTPUT - print all html lines */
#ifdef PRINTDEBUG
    printf("CHUNK OUTPUT CELLS:\n");
    for (i=0; i<chunk->numlines; i++)
        printf("%s\n", lines[i]);
    putchar('\n');
#endif

    return lines;
}

/* Function: free_lines
 * Frees up all memory associated with a string array, including the
 * array itself.
 */
void free_lines(int numlines, char** lines)
{
    int i;
    for (i=0; i<numlines; i++)
        free(lines[i]);
    free(lines);
}

/* Function: interlace_and_print_html
 * Takes two string arrays containing HTML code, and prints them into
 * outstream, interlacing them and adding all the necessary table code.
 */
void interlace_and_print_html(int numdeleted, char** deleted,
                              int numadded, char** added)
{
    int i;
    int max = numdeleted > numadded ? numdeleted : numadded;
    for (i=0; i<max; i++)
    {
        fputs(HINDT1 TABLE_ROWSTART NL HINDT2, outstream);
        if (i < numdeleted)
        {
            fputs(TABLE_CELL_DELETED TABLE_CELLSTART_DELETED, outstream);
            fputs(deleted[i], outstream);
            fputs(TABLE_CELLEND, outstream);
        }
        else
        {
            fputs(TABLE_CELLS_MISSING, outstream);
        }
        fputs(NL HINDT2, outstream);
        if (i < numadded)
        {
            fputs(TABLE_CELL_ADDED TABLE_CELLSTART_ADDED, outstream);
            fputs(added[i], outstream);
            fputs(TABLE_CELLEND, outstream);
        }
        else
        {
            fputs(TABLE_CELLS_MISSING, outstream);
        }
        fputs(NL HINDT1 TABLE_ROWEND NL, outstream);
    }
}

/* Function: get_unicode
 * Takes a reference to a string. Returns the Unicode character
 * representation of the first character in the string.
 * If this character is a UTF-8 multi-byte character, assumes UTF-8 format,
 * and increments *ptr so that *ptr now points to the last byte of the
 * character. If the character is invalid UTF-8, assumes ISO 8859-1 and
 * returns the single byte, not affecting *ptr.
 */
unsigned int get_unicode(unsigned char** ptr)
{
    unsigned char b1, b2, b3, b4;
    if (**ptr < 0x80)   /* ASCII or ASCII-equivalent */
    {}  /* Drop through to end */
    else if ((**ptr & 0xE0) == 0xC0)     /* 2-byte UTF-8 */
    {
        /* Check 2nd byte for UTF-8 conformance */
        if (((*ptr)[1] & 0xC0) == 0x80)
        {
            b1 = **ptr & 0x1F;
            (*ptr)++;   /* Move to 2nd byte */
            b2 = **ptr & 0x3F;
            return (b1 << 6) + b2;
        }
    }
    else if ((**ptr & 0xF0) == 0xE0)     /* 3-byte UTF-8 */
    {
        /* Check 2nd and 3rd byte */
        if (((*ptr)[1] & 0xC0) == 0x80 && ((*ptr)[2] & 0xC0) == 0x80)
        {
            b1 = **ptr & 0x0F;
            (*ptr)++;   /* Move to 2nd byte */
            b2 = **ptr & 0x3F;
            (*ptr)++;   /* Move to 3nd byte */
            b3 = **ptr & 0x3F;
            return (b1 << 12) + (b2 << 6) + b3;
        }
    }
    else if ((**ptr & 0xF8) == 0xF0)     /* 4-byte UTF-8 */
    {
        /* Check 2nd, 3rd and 4th byte */
        if (((*ptr)[1] & 0xC0) == 0x80 && ((*ptr)[2] & 0xC0) == 0x80
            && ((*ptr)[3] & 0xC0) == 0x80)
        {
            b1 = **ptr & 0x07;
            (*ptr)++;   /* Move to 2nd byte */
            b2 = **ptr & 0x3F;
            (*ptr)++;   /* Move to 3nd byte */
            b3 = **ptr & 0x3F;
            (*ptr)++;   /* Move to 4th byte */
            b4 = **ptr & 0x3F;
            return (b1 << 18) + (b2 << 12) + (b3 << 6) + b4;
        }
    }

    /* Assume it's ISO 8859-1 - ie. ASCII */
    return (unsigned int) **ptr;
}
