/* DIFFH - A diff frontend for output in HTML format.
 * Module:  INPUT
 * Author:  Matt Giuca
 * Created: 6/2/06
 * Updated: 15/4/06
 *
 * Copyright (C) 2006 Matt Giuca <mattgiuca@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either 
 * version 2 of the License, or (at your option) any later 
 * version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* Module: INPUT
 * This module reads from an input stream and outputs to UDiff structures.
 */

#include <ctype.h>
#include "global.h"
#include "input.h"

/* PRIVATE CONSTANTS */
#define CR 0x0D
#define LF 0x0A

/* Internal error codes */
#define ERR_INVALID_UDIFF 1
#define ERR_EMPTY_FILE 2
/* Error messages */
#define ERRMSG_INVALID_UDIFF "Input was not a valid unified diff file.\n" \
    "Use the -u or -Un arguments to the diff program to produce unified " \
    "diff."
#define ERRMSG_EMPTY_FILE "Input file was empty - possibly because the two" \
    " files match."

/* PRIVATE FUNCTION DECLARATIONS */
static void free_hunk(Hunk* hunk);
static char* read_input_line(FILE* instream);
static int get_file_info(char* line, char** filename, char** time,
    char** version);
static int parse_hunk_header(char* header, int* from_line_start,
    int* to_line_start);
static void print_error(int errcode);

/* MODULE FUNCTION CODE */

/* Function: read_diff_file
 * Takes an input stream, reads until the end-of-file, and returns a complete
 * UDiff structure with the file. Expects a universal diff file, or RCS/CVS
 * diff output, as a universal diff.
 * The return value is allocated memory and must be freed with free_udiff.
 * If errors are encountered, prints error message to stderr and returns
 * NULL.
 * Precond: instream must be a valid, readable stream.
 */
UDiff* read_diff_file(FILE* instream)
{
#ifdef PRINTDEBUG
    int i, j;
#endif
    int anysignificantlinesyet = FALSE;
    UDiff* diff = malloc(sizeof(UDiff));
    Hunk* thishunk;
    int diff_hunk_malloc_size;
    int hunk_line_malloc_size;
    char* line;
    size_t linelen;
    int from_line_start, to_line_start;
    char* this_hunk_line;
    char this_hunk_line_type;

    /* Set all to NULL so it can be free_udiff'd at any time */
    diff->from_filename = NULL;
    diff->from_time = NULL;
    diff->from_version = NULL;
    diff->to_filename = NULL;
    diff->to_time = NULL;
    diff->to_version = NULL;
    diff->hunk = malloc(DEFAULT_MALLOC * sizeof(Hunk));
    diff_hunk_malloc_size = DEFAULT_MALLOC;
    diff->numhunks = 0;

    while ((line = read_input_line(instream)) != NULL)
    {
        if (*line != '\0') anysignificantlinesyet = TRUE;
        /* Look for line beginning with "--- " */
        if (!strncmp(line, "--- ", 4))
        {
            if (*line == '\0'
                || !get_file_info(line+4, &diff->from_filename,
                &diff->from_time, &diff->from_version))
            {   /* Error */
                free_udiff(diff);
                free(line);
                return NULL;
            }
            free(line);
            break;
        }
        free(line);
    }
    if (diff->from_filename == NULL ||
        (line = read_input_line(instream)) == NULL ||
        strncmp(line, "+++ ", 4))   /* Next line must begin with "+++ " */
    {
        free_udiff(diff);
        print_error(anysignificantlinesyet ? ERR_INVALID_UDIFF :
            ERR_EMPTY_FILE);
        return NULL;
    }

    if (*line == '\0' || !get_file_info(line+4, &diff->to_filename,
        &diff->to_time, &diff->to_version))
    {   /* Error */
        free_udiff(diff);
        free(line);
        print_error(ERR_INVALID_UDIFF);
        return NULL;
    }
    free(line);

    /* Next line should commence hunks */
    line = read_input_line(instream);
    while (line != NULL)
    {   /* First line of each hunk should start and end with "@@" */
        linelen = strlen(line);
        if (linelen <= 6 || strncmp(line, "@@ ", 3)
            || strncmp(line+linelen-3, " @@", 3))
        {
            free_udiff(diff);
            print_error(ERR_INVALID_UDIFF);
            return NULL;
        }
        line[linelen-3] = '\0';     /* Terminate early */
        if (!parse_hunk_header(line+3, &from_line_start, &to_line_start))
        {
            free_udiff(diff);
            print_error(ERR_INVALID_UDIFF);
            return NULL;
        }

        /* Malloc and initialise hunk */
        thishunk = malloc(sizeof(Hunk));
        thishunk->from_line_start = from_line_start;
        thishunk->to_line_start = to_line_start;
        thishunk->line = malloc(DEFAULT_MALLOC * sizeof(char*));
        thishunk->line_type = malloc(DEFAULT_MALLOC * sizeof(char));
        hunk_line_malloc_size = DEFAULT_MALLOC;
        thishunk->numlines = 0;

        while ((line = read_input_line(instream)) != NULL)
        {
            if (*line == '\0')
            {   /* Only allowed at EOF
                    If any non-blank lines are found, error */
                do
                    line = read_input_line(instream);
                while (line != NULL && *line == '\0');
                if (line != NULL)
                {
                    free_udiff(diff);
                    free_hunk(thishunk);
                    print_error(ERR_INVALID_UDIFF);
                    return NULL;
                }
                break;
            }
            if (*line == '@')   /* New hunk */
                break;

            switch (*line)
            {
                case '-':
                this_hunk_line_type = LINE_DELETED;
                break;
                case '+':
                this_hunk_line_type = LINE_ADDED;
                break;
                case ' ':
                this_hunk_line_type = LINE_UNCHANGED;
                break;
                case '\\':
                continue;   /* Ignore this line, read another */
                break;
                default:
                free_udiff(diff);
                free_hunk(thishunk);
                print_error(ERR_INVALID_UDIFF);
                return NULL;
            }
            /* All clear - store the rest of the line in the hunk */
            this_hunk_line = malloc((strlen(line+1)+1) * sizeof(char));
            strcpy(this_hunk_line, line+1);
            if (thishunk->numlines >= hunk_line_malloc_size)
            {
                thishunk->line = realloc(thishunk->line,
                    (hunk_line_malloc_size*=2) * sizeof(char*));
                thishunk->line_type = realloc(thishunk->line_type,
                    (hunk_line_malloc_size) * sizeof(char));
            }
            thishunk->line[thishunk->numlines] = this_hunk_line;
            thishunk->line_type[thishunk->numlines] = this_hunk_line_type;
            thishunk->numlines++;
        }
        /* Store hunk in diff */
        if (diff->numhunks >= diff_hunk_malloc_size)
        {
            diff->hunk = realloc(diff->hunk,
                (diff_hunk_malloc_size*=2) * sizeof(Hunk*));
        }
        diff->hunk[diff->numhunks] = thishunk;
        diff->numhunks++;
    }

    /* DEBUG OUTPUT - print the whole UDiff structure */
#ifdef PRINTDEBUG
    printf("Diff from: filename = \"%s\", time = \"%s\", ver = \"%s\"\n",
        diff->from_filename, diff->from_time, diff->from_version);
    printf("Diff to: filename = \"%s\", time = \"%s\", ver = \"%s\"\n",
        diff->to_filename, diff->to_time, diff->to_version);
    for (i=0; i<diff->numhunks; i++)
    {
        printf("Hunk: Line %d; Line %d\n", diff->hunk[i]->from_line_start,
            diff->hunk[i]->to_line_start);
        for (j=0; j<diff->hunk[i]->numlines; j++)
        {
            putchar(diff->hunk[i]->line_type[j] == LINE_UNCHANGED ? ' ' :
                diff->hunk[i]->line_type[j] == LINE_DELETED ? '-' : '+');
            printf("%s\n", diff->hunk[i]->line[j]);
        }
    }
#endif
    return diff;
}

/* Function: free_udiff
 * Frees up all memory associated with a UDiff, including the diff itself.
 * This also frees all of the hunks linked to by the udiff.
 * Precond: The pointers in the structure must all point to memory allocated
 * with malloc, and must all be valid or NULL.
 */
void free_udiff(UDiff* diff)
{
    int i;

    free(diff->from_filename);
    free(diff->from_time);
    free(diff->from_version);
    free(diff->to_filename);
    free(diff->to_time);
    free(diff->to_version);
    for (i=0; i<diff->numhunks; i++)
        free_hunk(diff->hunk[i]);
    free(diff->hunk);
    free(diff);
}

/* Function: free_hunk
 * Frees up all memory associated with a hunk, including the diff itself.
 */
void free_hunk(Hunk* hunk)
{
    int i;
    for (i=0; i<hunk->numlines; i++)
        free(hunk->line[i]);
    free(hunk->line);
    free(hunk->line_type);
    free(hunk);
}

/* Function: read_input_line
 * Reads a line of any length from instream, and returns it as a char*.
 * The returned string will not contain a newline character.
 * The returned string is allocated memory and must be freed.
 * If instream is at the end, returns NULL.
 */
char* read_input_line(FILE* instream)
{
    char* line;
    size_t malloc_size = DEFAULT_MALLOC_BIG;
    size_t len = 0;
    int c;

    if (feof(instream)) return NULL;

    line = malloc(DEFAULT_MALLOC_BIG * sizeof(char));
    c = getc(instream);
    /* Note: Use CR and LF, not NL. NL is not guaranteed a value. */
    while (c != EOF && c != CR && c != LF)
    {
        if (len+1 >= malloc_size)
            line = realloc(line, (malloc_size*=2) * sizeof(char));
        line[len++] = (char) c;
        c = getc(instream);
    }
    if (c == CR)    /* If next char is LF, read that too */
    {
        c = getc(instream);
        if (c != LF) ungetc(c, instream);
    }
    line[len] = '\0';
    return line;
}

/* Function: get_file_info
 * Takes a line from the udiff file containing the tab-separated header
 * information for a source file. This line should NOT contain the - or +
 * signs at the beginning. Reads the line and stores the values in allocated
 * strings which are assigned to filename, time and version.
 * If the line is invalid, prints an error message and returns FALSE.
 * Otherwise returns TRUE.
 * Postcond: Upon returning false, filename, time and version will not have
 * been altered. Upon returning true, filename and time will have
 * been allocated; version may have been allocated or may not be altered.
 */
int get_file_info(char* line, char** filename, char** time, char** version)
{
    char* i;
    unsigned int filename_len = -1;
    char* time_start;
    unsigned int time_len;
    char* version_start;
    unsigned int version_len;

    for (i=line; *i; i++)
    {
        if (*i == '\t')
        {
            filename_len = (unsigned int) (i-line);
            break;
        }
    }
    if (filename_len == -1)  /* Didn't find tab */
    {
        print_error(ERR_INVALID_UDIFF);
        return FALSE;
    }
    /* Now guaranteed to finish - we can allocate memory */
    *filename = malloc((filename_len + 1) * sizeof(char));
    strncpy(*filename, line, filename_len);
    (*filename)[filename_len] = '\0';

    time_start = line + filename_len + 1;
    for (i=time_start; *i && *i != '\t'; i++);
    time_len = (unsigned int) (i-time_start);
    *time = malloc((time_len + 1) * sizeof(char));
    strncpy(*time, time_start, time_len);
    (*time)[time_len] = '\0';

    if (*i == '\t') /* There is a version here too */
    {
        version_start = time_start + time_len + 1;
        for (i=version_start; *i; i++);
        version_len = (unsigned int) (i-version_start);
        *version = malloc((version_len + 1) * sizeof(char));
        strncpy(*version, version_start, version_len);
        (*version)[version_len] = '\0';
    }

    return TRUE;
}

/* Function: parse_hunk_header
 * Takes a line from the udiff file containing the hunk header line.
 * This line should NOT contain the @@ or spaces around the edge.
 * Reads the line and stores the values in allocated ints which are assigned
 * to from_line_start and to_line_start.
 * If the line is invalid returns FALSE.
 * Otherwise returns TRUE.
 */
int parse_hunk_header(char* header, int* from_line_start, int* to_line_start)
{
    char* i;
    int fls = 0, tls = 0;
    if (*header != '-') return FALSE;
    for (i=header+1; *i && *i != ',' && *i != ' '; i++)
    {   /* Construct fls from the digits */
        if (!isdigit(*i)) return FALSE;
        fls = (fls * 10) + (*i-'0');
    }
    if (*i == '\0') return FALSE;

    /* Ignore 2nd number, look for space */
    if (*i != ' ')     /* If hit a ' ', there was no comma, don't worry */
    {
        i++;
        for (; *i && *i != ' '; i++)
            if (!isdigit(*i)) return FALSE;
        if (*i == '\0') return FALSE;
    }
    i++;
    if (*i != '+') return FALSE;

    for (i++; *i && *i != ','; i++)
    {   /* Construct tls from the digits */
        if (!isdigit(*i)) return FALSE;
        tls = (tls * 10) + (*i-'0');
    }

    /* See it out - must end with a number then that's it */
    if (*i != '\0')     /* If hit a \0, there was no comma, don't worry */
    {
        i++;
        for (; *i ; i++)
            if (!isdigit(*i)) return FALSE;
    }

    *from_line_start = fls;
    *to_line_start = tls;
    return TRUE;
}

/* Function: print_error
 * Prints out an error message with the specified error code
 */
void print_error(int errcode)
{
    switch (errcode)
    {
        case ERR_INVALID_UDIFF:
        fputs(ERRMSG_INVALID_UDIFF NL, stderr);
        break;
        case ERR_EMPTY_FILE:
        fputs(ERRMSG_EMPTY_FILE NL, stderr);
        break;
    }
}
